from astropy import units
from astropy.coordinates import SkyCoord
from astropy.coordinates.name_resolve import NameResolveError

from octans import Sky
from unittest import TestCase

from octans.errors import ObjectNotFoundError


class TestSky(TestCase):

    def test_create(self):
        xy_leo = Sky("xy leo")
        sky_coord = SkyCoord.from_name("xy Leo")
        self.assertEqual(xy_leo.skycoord, sky_coord)

    def test_create_from_coordinates(self):
        sky_coord = SkyCoord.from_name("xy Leo")
        xy_leo = Sky.from_coordinates(sky_coord.ra, sky_coord.dec)
        self.assertIn("xy leo", xy_leo.name.lower())

    def test_create_does_not_exist(self):
        with self.assertRaises(NameResolveError):
            _ = Sky("WA Dra")

    def test_create_from_coordinates_does_not_exist(self):
        with self.assertRaises(ObjectNotFoundError):
            _ = Sky.from_coordinates(1, 0, radius=1 * units.arcsec)

    def test_change_name(self):
        xy_leo = Sky("xy leo")
        sky_coord = SkyCoord.from_name("xy Leo")
        self.assertEqual(xy_leo.skycoord.ra, sky_coord.ra)
        self.assertEqual(xy_leo.skycoord.dec, sky_coord.dec)

        xy_leo.name = "AW Dra"
        sky_coord_aw_dra = SkyCoord.from_name("AW DRA")
        self.assertEqual(xy_leo.skycoord.ra, sky_coord_aw_dra.ra)
        self.assertEqual(xy_leo.skycoord.dec, sky_coord_aw_dra.dec)

    def test_change_sky_coord(self):
        xy_leo = Sky("xy leo")
        sky_coord = SkyCoord.from_name("xy Leo")
        self.assertEqual(xy_leo.skycoord.ra, sky_coord.ra)
        self.assertEqual(xy_leo.skycoord.dec, sky_coord.dec)

        sky_coord_aw_dra = SkyCoord.from_name("AW DRA")

        with self.assertRaises(AttributeError):
            xy_leo.skycoord = sky_coord_aw_dra



from abc import ABC, abstractmethod
from logging import Logger
from typing import Self, Optional, Literal, Iterator

from astropy import units

from . import Sky, XLightCurve
from .utils import NAngleType, Minima


class ModelPortal(ABC):
    @classmethod
    @abstractmethod
    def from_name(cls, name: str) -> Self:
        """Create a portal object from a given name"""

    @classmethod
    @abstractmethod
    def from_coordinates(cls, ra: NAngleType, dec: NAngleType, radius: NAngleType = 2 * units.arcmin) -> Self:
        """Create a portal object from a given coordinates"""

    @property
    @abstractmethod
    def sky(self) -> Sky:
        """Return the sky object related to the portal"""

    @sky.setter
    @abstractmethod
    def sky(self, sky: Sky) -> None:
        """Prevents setting the sky value"""

    @abstractmethod
    def kt(self, mission: Optional[Literal['kepler', 'tess']] = None) -> Iterator[XLightCurve]:
        """Retrieves either kepler, k2 or tess lightcurve from server"""

    @abstractmethod
    def kepler(self) -> Iterator[XLightCurve]:
        """Retrieves kepler lightcurve from server"""

    @abstractmethod
    def tess(self) -> Iterator[XLightCurve]:
        """Retrieves tess lightcurve from server"""

    @abstractmethod
    def asas(self) -> Iterator[XLightCurve]:
        """Retrieves ASAS lightcurve from server"""

    @abstractmethod
    def var_astro(self) -> Minima:
        """Returns minimas from VarAstro"""

    @abstractmethod
    def oc_gateway(self) -> Minima:
        """Returns minimas from VarAstro"""

    @abstractmethod
    def etd(self) -> Minima:
        """Returns minimas from VarAstro ETD"""


class ModelASAS(ABC):
    @classmethod
    @abstractmethod
    def from_sky(cls, sky: Sky, max_dist: NAngleType = 20 * units.arcsec, time_out: NAngleType = 3 * units.s) -> Self:
        """Retrieves the asas_id of the given coordinates"""

    @abstractmethod
    def data(self) -> str:
        """Retrieves the raw lightcurve"""

    @abstractmethod
    def get(self) -> Iterator[XLightCurve]:
        """Retrieves the XLightCurve """


class ModelVarAstro(ABC):
    @classmethod
    @abstractmethod
    def from_sky(cls, sky: Sky, max_dist: NAngleType = 20 * units.arcmin, time_out: NAngleType = 3 * units.s,
                 logger: Optional[Logger] = None) -> Self:
        """Creates a VarAstro object from a given coordinates"""

    @abstractmethod
    def get(self) -> Minima:
        """Retrieves the minima times from VarAstro"""



from logging import Logger, getLogger
from typing import Optional

from typing_extensions import Self

from .errors import ObjectNotFoundError
from .utils import NAngleType, unit_checker
from .utils import degree_checker
from .model_sky import ModelSky

from astropy import units
from astropy.coordinates import SkyCoord

from astroquery.simbad import Simbad


class Sky(ModelSky):
    def __init__(self, name: str, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.__name = name
        self.__skycoord = self.resolve

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(name: {self.name})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.__name}')"

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name
        self.__skycoord = self.resolve

    @property
    def skycoord(self) -> SkyCoord:
        return self.__skycoord

    @skycoord.setter
    def skycoord(self, skycoord: SkyCoord) -> None:
        self.logger.error("This attribute is immutable and cannot be changed.")
        raise AttributeError("This attribute is immutable and cannot be changed.")

    @property
    def resolve(self) -> SkyCoord:
        return SkyCoord.from_name(self.__name)

    @classmethod
    def from_coordinates(cls, ra: NAngleType, dec: NAngleType, radius: NAngleType = 2 * units.arcmin,
                         logger: Optional[Logger] = None) -> Self:
        if logger is None:
            logger = getLogger(__name__)

        the_radius = unit_checker(radius, units.arcsec)
        the_ra = degree_checker(ra)
        the_dec = degree_checker(dec)
        coord = SkyCoord(ra=the_ra, dec=the_dec)
        result_table = Simbad.query_region(coord, radius=the_radius)

        if result_table is None:
            logger.error("No objects found.")
            raise ObjectNotFoundError("No objects found.")

        sky_coords = SkyCoord(result_table["RA"], result_table["DEC"], unit=(units.hourangle, units.deg))
        closest_ids, closest_dists, _ = coord.match_to_catalog_sky(sky_coords)
        return cls(result_table[closest_ids]["MAIN_ID"], logger=logger)



from unittest import TestCase

from octans import Sky, NASAExoplanetArchive
from octans.errors import NoNASAExoplanetArchiveObjectError


class TestNasaExoplanetArchive(TestCase):
    def test_create(self):
        wasp_3_b = Sky("wasp-3 b")
        _ = NASAExoplanetArchive(wasp_3_b)

    def test_resolve(self):
        wasp_3_b = Sky("wasp-3 b")
        exoplanet = NASAExoplanetArchive(wasp_3_b)
        for i in ["pl_name", "pl_orbper", "pl_orbpererr1", "pl_orbpererr2"]:
            self.assertIn(i, exoplanet.resolve().columns)

    def test_period(self):
        wasp_3_b = Sky("wasp-3 b")
        exoplanet = NASAExoplanetArchive(wasp_3_b)
        for i in ['name', 'P', 'PEmin', 'PEmax']:
            self.assertIn(i, exoplanet.period().columns)

    def test_create_does_not_Exist(self):
        xy_leo = Sky("xy leo")
        exoplanet = NASAExoplanetArchive(xy_leo)

        with self.assertRaises(NoNASAExoplanetArchiveObjectError):
            _ = exoplanet.resolve()



from logging import getLogger, Logger
from typing import Self, Optional, Literal, Iterator, List, Any

import numpy as np
import pandas as pd
from astropy import units
import lightkurve as lk
from astropy.coordinates import SkyCoord
from astropy.time import Time, TimeDelta

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.service import Service as ChromeService
from webdriver_manager.chrome import ChromeDriverManager

from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from selenium.common.exceptions import TimeoutException

from bs4 import BeautifulSoup

import requests

from . import XLightCurve
from .errors import NoLightCurveError, PageNotFoundError, NoASASObjectError, NoVarAstroObjectError, \
    NoDataFoundError
from .model_portal import ModelPortal, ModelASAS, ModelVarAstro
from .sky import Sky
from .utils import NAngleType, degree_checker, time_checker, Minima

ASAS_XPATHS = {
    "input_coordinate": "/html/body/div[2]/form/table/tbody/tr[2]/td[1]/input",
    "input_radius": "/html/body/div[2]/form/table/tbody/tr[2]/td[2]/input",
    "input_arcsec_radio": "/html/body/div[2]/form/table/tbody/tr[2]/td[3]/input[1]",
    "input_show": "/html/body/div[2]/form/table/tbody/tr[2]/td[5]/input[1]",
    "input_results": "/html/body/div[2]/form/table/tbody/tr[1]/th[5]/input",
    "result_table": "/html/body/div[2]/table"
}

VarAstro_XPATHS = {
    "result_table_first_row": "/html/body/div[2]/div[2]/div[3]/div[2]/div[2]/div[2]/div/div/div/div/table/tbody/tr[1]"
}


def check_exists_by_xpath(driver, xpath, timeout=3):
    """
    Check if an element does exist in a webpage
    """
    try:
        _ = WebDriverWait(driver, timeout).until(EC.presence_of_element_located((By.XPATH, xpath)))
        return True
    except TimeoutException:
        return False


def line_formatter(line) -> List[Any]:
    """
    Format the lines obtained form ASAS's web site
    """
    the_line = list(map(float, line[:11])) + [line[11], int(line[12])]
    the_line[0] += 2450000
    return the_line


def magnitude_to_flux(magnitude: float, ref_magnitude: float = 0.0, ref_flux: float = 1.0):
    """
    Convert Magnitude to Flux
    """
    flux_density = np.power(
        10,
        (magnitude - ref_magnitude) / - 2.5
    ) * ref_flux / (units.electron * units.second)
    return flux_density.decompose()


class Portal(ModelPortal):
    def __init__(self, sky: Sky, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.__sky = sky

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(sky: {self.sky})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.sky})"

    @classmethod
    def from_name(cls, name: str, logger: Optional[Logger] = None) -> Self:
        return cls(Sky(name), logger=logger)

    @classmethod
    def from_coordinates(cls, ra: NAngleType, dec: NAngleType, radius: NAngleType = 2 * units.arcmin,
                         logger: Optional[Logger] = None) -> Self:
        return cls(Sky.from_coordinates(ra, dec, radius), logger=logger)

    @property
    def sky(self) -> Sky:
        return self.__sky

    @sky.setter
    def sky(self, sky: Sky) -> None:
        self.logger.error("This attribute is immutable and cannot be changed.")
        raise AttributeError("This attribute is immutable and cannot be changed.")

    def kt(self, mission: Optional[Literal['kepler', 'tess']] = None) -> Iterator[XLightCurve]:
        search_result = lk.search_lightcurve(self.sky.name, mission=mission)

        if len(search_result) == 0:
            self.logger.error(f"No data found for target {self.sky.name} at {mission}.")
            raise NoLightCurveError(f"No data found for target {self.sky.name} at {mission}.")

        lc_collection = search_result.download_all()
        for lc in lc_collection:
            yield XLightCurve(time=lc.time.jd, flux=lc.flux.value, flux_err=lc.flux_err.value)

    def kepler(self) -> Iterator[XLightCurve]:
        return self.kt("kepler")

    def tess(self) -> Iterator[XLightCurve]:
        return self.kt("tess")

    def asas(self) -> Iterator[XLightCurve]:
        asas = ASAS.from_sky(self.sky)
        return asas.get()

    def var_astro(self) -> Minima:
        var_astro = VarAstro.from_sky(self.sky)
        return var_astro.get()

    def oc_gateway(self) -> Minima:
        return self.var_astro()

    def etd(self) -> Minima:
        etd = ETD.from_sky(self.sky)
        return etd.get()


class ASAS(ModelASAS):
    def __init__(self, asas_id: str, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.asas_id = asas_id

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(asas_id: {self.asas_id})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.asas_id}')"

    @classmethod
    def from_sky(cls, sky: Sky, max_dist: NAngleType = 20 * units.arcsec, time_out: NAngleType = 3 * units.s,
                 logger: Optional[Logger] = None) -> Self:
        if logger is None:
            logger = getLogger(__name__)

        the_max_dist = degree_checker(max_dist)
        the_time_out = time_checker(time_out)

        options = Options()
        options.add_argument("--headless")
        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), options=options)
        driver.get("https://www.astrouw.edu.pl/asas/?page=acvs")

        if not check_exists_by_xpath(driver, ASAS_XPATHS["input_coordinate"], timeout=the_time_out.to(units.s).value):
            logger.error("Page Not Found")
            raise PageNotFoundError("Page Not Found")

        # Write the coordinates
        coordinates_box = driver.find_element(By.XPATH, ASAS_XPATHS["input_coordinate"])
        coordinates_box.clear()

        coordinates_box.send_keys(f"{sky.skycoord.ra.to_string('hour', sep=':', precision=0)} "
                                  f"{sky.skycoord.dec.to_string('deg', sep=':', precision=0)}")

        # Clear the radius box and write the actual value
        radius_box = driver.find_element(By.XPATH, ASAS_XPATHS["input_radius"])
        radius_box.clear()
        radius_box.send_keys(f"{int(the_max_dist.to(units.arcsec).value)}")

        # Make sure the radius is in arcseconds
        arcsec_radio = driver.find_element(By.XPATH, ASAS_XPATHS["input_arcsec_radio"])
        arcsec_radio.click()

        # Show all results
        all_radio = driver.find_element(By.XPATH, ASAS_XPATHS["input_show"])
        all_radio.click()

        # Get the results
        result_button = driver.find_element(By.XPATH, ASAS_XPATHS["input_results"])
        result_button.click()

        if not check_exists_by_xpath(driver, ASAS_XPATHS["result_table"], timeout=the_time_out.to(units.s).value):
            logger.error("There's no such ASAS object")
            raise NoASASObjectError("There's no such ASAS object")

        page_html = driver.page_source
        page_soup = BeautifulSoup(page_html, 'html.parser')

        table_soup = page_soup.select_one("table")
        if len(table_soup) == 0:
            logger.error("There's no such ASAS object")
            raise NoASASObjectError("There's no such ASAS object")

        trs_soup = table_soup.find_all("tr", {"align": "center"})

        if len(trs_soup) == 0:
            logger.error("There's no such ASAS object")
            raise NoASASObjectError("There's no such ASAS object")

        min_distance = 180 * units.deg
        best_match = None

        for tr_soup in trs_soup:
            tds_soup = tr_soup.find_all("td")
            if len(tds_soup) != 10:
                logger.warning("The table is not found")
                continue

            each_sky_coord = SkyCoord(
                ra=tds_soup[1].text.strip(),
                dec=tds_soup[2].text.strip(),
                unit=("hourangle", "degree")
            )
            each_separation = each_sky_coord.separation(sky.skycoord)

            if each_separation < the_max_dist:
                if each_separation < min_distance:
                    min_distance = each_separation
                    best_match = tds_soup[0]

        if best_match is None:
            logger.error("There's no such ASAS object")
            raise NoASASObjectError("There's no such ASAS object")

        return cls(best_match.text.strip(), logger=logger)

    def data(self) -> str:
        respond = requests.get(f"https://www.astrouw.edu.pl/cgi-asas/asas_cgi_get_data?{self.asas_id},asas3")
        if respond.status_code != 200:
            self.logger.error("There's no such ASAS object")
            raise NoASASObjectError("There's no such ASAS object")

        return str(respond.text)

    def get(self) -> Iterator[XLightCurve]:
        content = self.data()
        the_lc = pd.DataFrame(
            columns=["HJD", "MAG_4", "MAG_0", "MAG_1", "MAG_2", "MAG_3", "MER_4", "MER_0", "MER_1", "MER_2", "MER_3",
                     "GRADE", "FRAME"]
        )

        chunks = list(map(lambda x: "#ndata=" + x, content.split("#ndata=")[1:]))
        for chunk in chunks:
            header = None
            data = []
            for line in chunk.split("\n"):
                if line.startswith("#"):
                    if "HJD" in line:
                        header = line.replace("#", "").strip().split()
                else:
                    if line:
                        data.append(line_formatter(line.split()))

            the_lc = pd.concat([
                the_lc,
                pd.DataFrame(
                    data=data,
                    columns=header
                )
            ])
        the_lc = the_lc.sort_values(by=["HJD"])
        for extension in range(5):
            yield XLightCurve(
                Time(the_lc["HJD"], format="jd", scale="utc"),
                magnitude_to_flux(the_lc[f"MAG_{extension}"].to_numpy()),
                magnitude_to_flux(the_lc[f"MER_{extension}"].to_numpy())
            )


class VarAstro(ModelVarAstro):
    def __init__(self, var_astro_id: int, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.var_astro_id = var_astro_id

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(var_astro_id: {self.var_astro_id})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.var_astro_id}')"

    @classmethod
    def from_sky(cls, sky: Sky, max_dist: NAngleType = 20 * units.arcmin, time_out: NAngleType = 3 * units.s,
                 logger: Optional[Logger] = None) -> Self:
        if logger is None:
            logger = getLogger(__name__)

        the_max_dist = degree_checker(max_dist)
        the_time_out = time_checker(time_out)

        options = Options()
        options.add_argument("--headless")
        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), options=options)
        driver.get(
            f"https://var.astro.cz/en/Stars?pageId=1&pageSize=20&ra="
            f"{sky.skycoord.ra.degree}&dec={sky.skycoord.dec.degree}&radiusArcmin="
            f"{int(the_max_dist.to(units.arcmin).value)}")

        WebDriverWait(driver, the_time_out.to(units.s).value).until(
            EC.presence_of_element_located((By.XPATH, VarAstro_XPATHS["result_table_first_row"]))
        )

        rows = driver.find_elements(By.XPATH, VarAstro_XPATHS["result_table_first_row"])

        data = rows[0].find_elements(By.TAG_NAME, "td")
        if data[0].text.strip() == "No data available in table":
            logger.error("There's no such VarAstro object")
            raise NoVarAstroObjectError("There's no such VarAstro object")

        if data[1].text.strip() == "0":
            logger.error("There's no entry for the given object in VarAstro")
            raise NoLightCurveError("There's no entry for the given object in VarAstro")

        var_astro_id = int(data[0].text.strip())

        driver.close()
        return cls(var_astro_id, logger=logger)

    def get(self) -> Minima:
        options = Options()
        options.add_argument("--headless")
        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), options=options)
        driver.get(
            f"https://var.astro.cz/en/Stars/{self.var_astro_id}"
        )

        csv_data = driver.execute_script(f"""
                return varastro.shared.fetch("/api/charts/ocdiagram/{self.var_astro_id}?v=0").then(function(n) {{
                    if (n.status === 200) return n.json();
                    return n;
                }}).then(function(t) {{
                    return t;
                }})
                """)

        if csv_data is None:
            raise NoDataFoundError("No data found for target.")

        if "dataPrimary" not in csv_data.keys():
            raise NoVarAstroObjectError("There's no such VarAstro object")

        time = []
        time_error = []
        for data in csv_data["dataPrimary"]:
            time.append(float(data["jd"]))
            time_error.append(np.nan)

        return Minima(Time(time, format="jd", scale="utc"), TimeDelta(time_error, format="jd"))


class ETD(ModelVarAstro):
    def __init__(self, etd_id: int, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.etd_id = etd_id

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(etd_id: {self.etd_id})"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.etd_id}')"

    @classmethod
    def from_sky(cls, sky: Sky, max_dist: NAngleType = 20 * units.arcmin, time_out: NAngleType = 3 * units.s,
                 logger: Optional[Logger] = None) -> Self:
        if logger is None:
            logger = getLogger(__name__)

        the_max_dist = degree_checker(max_dist)
        the_time_out = time_checker(time_out)

        options = Options()
        options.add_argument("--headless")
        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), options=options)
        driver.get(
            f"https://var.astro.cz/en/Exoplanets?pageId=1&pageSize=20&ra="
            f"{sky.skycoord.ra.degree}&dec={sky.skycoord.dec.degree}&radiusArcmin="
            f"{int(the_max_dist.to(units.arcmin).value)}"
        )

        WebDriverWait(driver, the_time_out.to(units.s).value).until(
            EC.presence_of_element_located((By.XPATH, VarAstro_XPATHS["result_table_first_row"]))
        )

        rows = driver.find_elements(By.XPATH, VarAstro_XPATHS["result_table_first_row"])

        data = rows[0].find_elements(By.TAG_NAME, "td")
        if data[0].text.strip() == "No data available in table":
            logger.error("There's no such ETD object")
            raise NoVarAstroObjectError("There's no such ETD object")

        if data[1].text.strip() == "0":
            logger.error("There's no entry for the given object in ETD")
            raise NoLightCurveError("There's no entry for the given object in ETD")

        var_astro_id = int(data[0].text.strip())

        driver.close()
        return cls(var_astro_id, logger=logger)

    def get(self) -> Minima:
        options = Options()
        options.add_argument("--headless")
        driver = webdriver.Chrome(service=ChromeService(ChromeDriverManager().install()), options=options)
        driver.get(
            f"https://var.astro.cz/en/Exoplanets/{self.etd_id}"
        )

        csv_data = driver.execute_script(f"""
                return varastro.shared.fetch("/api/charts/exottvdiagram/{self.etd_id}?v=0").then(function(n) {{
                    if (n.status === 200) return n.json();
                    return n;
                }}).then(function(t) {{
                    return t;
                }})
                """)

        if csv_data is None:
            raise NoDataFoundError("No data found for target.")

        if "data" not in csv_data.keys():
            raise NoVarAstroObjectError("There's no such ETD object")

        time = []
        time_error = []
        for data in csv_data["data"]:
            time.append(float(data["jd"]))
            time_error.append(np.nan)

        return Minima(Time(time, format="jd", scale="utc"), TimeDelta(time_error, format="jd"))


class Utilities:
    def __init__(self, sky: Sky, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.sky = sky

    def period(self):
        pass



from abc import ABC, abstractmethod

import pandas as pd


class ExoplanetEUModel(ABC):
    @abstractmethod
    def resolve(self) -> pd.DataFrame:
        """Resolves the given sky coordinate to the database"""

    @abstractmethod
    def period(self) -> pd.DataFrame:
        """Returns the period of the exoplanet"""


class NASAExoplanetArchiveModel(ABC):
    @abstractmethod
    def resolve(self) -> pd.DataFrame:
        """Resolves the given sky coordinate to the database"""

    @abstractmethod
    def period(self) -> pd.DataFrame:
        """Returns the period of the exoplanet"""


class PeriodModel:
    @abstractmethod
    def nasa_exoplanet_archive(self) -> pd.DataFrame:
        """Returns the period of the exoplanet from Nasa Exoplanet Archive"""

    @abstractmethod
    def exoplanet_eu(self) -> pd.DataFrame:
        """Returns the period of the exoplanet from Exoplanet EU"""

    def all(self) -> pd.DataFrame:
        """Returns the period of the exoplanet from all available databases"""



import platform
from dataclasses import dataclass
from logging import Logger, getLogger
from pathlib import Path
from typing import Union, Optional, Iterator

from typing_extensions import Self

import numpy as np
from astropy import units
from astropy.time import Time, TimeDelta

Numeric = Union[int, float]
AngleType = units.Quantity["angle"]
TimeType = units.Quantity["time"]
NAngleType = Union[Numeric, units.Quantity["angle"]]
NTimeType = Union[Numeric, units.Quantity["time"]]


@dataclass
class Boundaries:
    lower: Time
    upper: Time
    logger: Optional[Logger] = None

    def __str__(self) -> str:
        if self.lower.isscalar:
            return f"[{self.lower.jd}, {self.upper.jd}]"

        return str([each for each in self])

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        return len(self.lower)

    def __iter__(self) -> Iterator[Self]:
        for lower, upper in zip(self.lower, self.upper):
            yield self.__class__(lower, upper)

    def __post_init__(self) -> None:

        if self.logger is None:
            self.logger = getLogger(__name__)

        if self.lower.isscalar:
            if not self.upper.isscalar:
                self.logger.error("Number of Lower and Upper bounds must match")
                raise ValueError("Number of Lower and Upper bounds must match")

        else:
            if len(self.lower) != len(self.upper):
                self.logger.error("Number of Lower and Upper bounds must match")
                raise ValueError("Number of Lower and Upper bounds must match")

    def __getitem__(self, key: Union[int, slice]) -> Self:
        return self.__class__(self.lower[key], self.upper[key])


@dataclass
class Minima:
    time: Time
    error: TimeDelta
    logger: Optional[Logger] = None

    def __str__(self) -> str:
        if self.time.isscalar:
            return f"{self.time.jd} ± {self.error.jd}"

        return str([each for each in self])

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        return len(self.time)

    def __iter__(self) -> Iterator[Self]:
        for time, error in zip(self.time, self.error):
            yield self.__class__(time, error)

    def __post_init__(self) -> None:
        if self.logger is None:
            self.logger = getLogger(__name__)

        if self.time.isscalar:
            if not self.error.isscalar:
                self.logger.error("Length of time must be equal to length of error")
                raise ValueError("Length of time must be equal to length of error")

        else:
            if len(self.time) != len(self.error):
                self.logger.error("Length of time must be equal to length of error")
                raise ValueError("Length of time must be equal to length of error")

    def __getitem__(self, key: Union[int, slice]) -> Self:
        return self.__class__(self.time[key], self.error[key])


def normalize(value: float, actual_min: float, actual_max: float, to_min=0.0, to_max=1.0) -> float:
    flip = actual_min > actual_max
    actual_min, actual_max = min(actual_min, actual_max), max(actual_min, actual_max)
    to_min, to_max = min(to_min, to_max), max(to_min, to_max)

    normalized = (value - actual_min) / (actual_max - actual_min)

    if flip:
        normalized = 1.0 - normalized

    return float(to_min + normalized * (to_max - to_min))


def interpolate(value: float, start_value: float, end_value: float) -> float:
    start_value, end_value = min(start_value, end_value), max(start_value, end_value)
    return start_value + (end_value - start_value) * value


def neighbor(array, value):
    return (np.abs(array - value)).argmin()


def unit_checker(value, unit: units.Quantity) -> units.Quantity:
    if isinstance(value, (float, int)):
        return value * unit

    if unit.is_equivalent(value):
        return value
    else:
        raise ValueError("Bad unit")


def degree_checker(value, unit: units.Quantity = units.deg) -> units.Quantity:
    return unit_checker(value, unit)


def time_checker(value, unit: units.Quantity = units.s) -> units.Quantity:
    return unit_checker(value, unit)


def database_dir():
    home_dir = Path.home()
    if platform.system() == "Windows":
        settings_dir = home_dir / "AppData" / "Local" / "octans"
    elif platform.system() == "Darwin":
        settings_dir = home_dir / "Library" / "Application Support" / "octans"
    elif platform.system() == "Linux":
        settings_dir = home_dir / ".config" / "octans"
    else:
        settings_dir = home_dir / "octans"

    if not settings_dir.exists():
        settings_dir.mkdir()

    return settings_dir



from abc import ABC, abstractmethod
from typing_extensions import Self

from astropy import units
from astropy.coordinates import SkyCoord

from .utils import NAngleType


class ModelSky(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        """Return the name of the sky object"""

    @property
    @abstractmethod
    def skycoord(self) -> SkyCoord:
        """Return the SkyCoord of the sky object"""

    @property
    @abstractmethod
    def resolve(self) -> SkyCoord:
        """Resolves the name of the sky object"""

    @classmethod
    @abstractmethod
    def from_coordinates(cls, ra: NAngleType, dec: NAngleType, radius: NAngleType = 2 * units.arcmin) -> Self:
        """Creates a Sky object from a given coordinate"""



from logging import Logger

import numpy as np
from logging import getLogger
from typing import Optional
from scipy.optimize import curve_fit


# https://github.com/hdeeg/KvW
def kvw(time, flux, init_minflux=False, rms=None, nfold=5, notimeoff=False, debug=0, logger: Optional[Logger] = None):
    if logger is None:
        logger = getLogger(__name__)

    errflag = 0
    npts = len(time)
    minid = npts // 2

    time0 = 0
    if not notimeoff:
        time0 = np.floor(time[minid] * 10) / 10
        time -= time0

    if debug >= 3:
        for i in range(npts):
            logger.info(i, flux[i])

    difftime = time[1:] - time[:-1]
    mediandiff = np.median(difftime)
    if np.max(difftime) / mediandiff >= 1.01 or mediandiff / np.min(difftime) >= 1.01:
        logger.warning('KVW: WARN: Time-points not equidistant within 1%. Raised errorflag')
        errflag += 1

    if init_minflux:
        minid = int(np.argmin(flux))
        if debug >= 2:
            logger.info('minid from init_minflux', minid)

    if debug >= 1:
        logger.info('minid ', minid)

    noffr = (nfold - 1) / 4.0
    noff = int(noffr)
    minfoldid = minid - noff
    maxfoldid = minid + noff
    nleft = minfoldid
    nright = npts - maxfoldid - 1
    z = min(nleft, nright)
    minpid = minfoldid - z
    maxpid = maxfoldid + z

    if debug >= 2:
        logger.info('points considered in pairings:', minpid, ' to ', maxpid)

    if debug >= 1:
        logger.info('Z: ', z)

    if z < 3:
        raise ValueError(
            'Error: Less than 3 points in in/egress can be paired. Decrease nfold parameter or provide more datapoints for in- or egress')

    s = np.zeros(nfold)
    foldidf = np.zeros(nfold)
    for segid in range(nfold):
        foldidf[segid] = minid - noffr + segid / 2.0
        foldidi = int(foldidf[segid] + 0.0001)
        if abs(foldidf[segid] - foldidi) <= 0.01:
            for i in range(1, z + 1):
                idlo = foldidi - i
                idhi = foldidi + i
                s[segid] += (flux[idlo] - flux[idhi]) ** 2
        else:
            for i in range(1, z + 1):
                idlo = foldidi - i + 1
                idhi = foldidi + i
                s[segid] += (flux[idlo] - flux[idhi]) ** 2

    minidf = int(np.floor(foldidf[0]))
    maxidf = int(np.floor(foldidf[-1] + 0.50001))
    popt, _ = curve_fit(lambda x, a, b: a + b * x, np.arange(minidf, maxidf + 1), time[minidf:maxidf + 1])
    if debug >= 2:
        logger.info('popt: ', popt)

    timef = popt[0] + popt[1] * foldidf
    if debug >= 2:
        logger.info('timef (orig): ', timef)

    if nfold >= 5:
        minSid = np.argmin(s)
        nSleft = minSid
        nSright = nfold - minSid - 1

        if nSleft - nSright >= 2:  # More than 2 more points on left than on right, cut left points
            timef = timef[nSleft - nSright - 1:]
            s = s[nSleft - nSright - 1:]

        if nSright - nSleft >= 2:  # More than 2 more points on right than on left, cut right points
            timef = timef[:nSleft - nSright + 1]
            s = s[:nSleft - nSright + 1]
        if debug >= 2:
            logger.info('nSleft: ', nSleft, ' nSright: ', nSright)

    if debug >= 2:
        logger.info('foldidf ', foldidf)
        logger.info('timef ', timef)
        logger.info('     S ', s)

    cp = np.polyfit(timef, s, 2, full=False)
    if debug >= 1:
        logger.info('cp (kvw): ', cp)

    mintim = -cp[1] / (2.0 * cp[0]) + time0
    kvwminerr = np.sqrt((4 * cp[0] * cp[2] - cp[1] ** 2) / (4 * cp[0] ** 2 * (z - 1)))

    if rms is None:
        rms = np.sqrt(np.min(s) / (2 * (z - 1)))

    cp[2] = (z - 1) * 2 * rms ** 2 + cp[1] ** 2 / (4 * cp[0])
    minerr = np.sqrt((4 * cp[0] * cp[2] - cp[1] ** 2) / (4 * cp[0] ** 2 * (z - 1)))
    if debug >= 1:
        logger.info('cp (revised): ', cp)
        logger.info('type of cp:', type(cp))
        logger.info('rms:', rms)
        logger.info('minerr: ', minerr)

    return mintim, minerr, kvwminerr, errflag



from unittest import TestCase

from octans import XLightCurve
import numpy as np

np.random.seed(42)


class TestXLightCurve(TestCase):
    def setUp(self):
        self.time = np.linspace(2460619.0, 2460619.30, 100)
        self.flux = np.sin(self.time * np.pi * 20)
        self.flux[self.flux > 0.5] = self.flux.max()
        self.flux += 1.1

        self.mean = self.flux.mean() / 2
        self.std = self.flux.std() / 10

        noise = np.random.normal(self.mean, self.std, self.flux.shape)
        self.flux += noise

    def test_create(self):
        _ = XLightCurve(time=self.time.tolist(), flux=self.flux.tolist())

    def test_boundaries(self):
        xlc = XLightCurve(time=self.time.tolist(), flux=self.flux.tolist())
        boundaries = xlc.boundaries_extrema()
        self.assertEqual(len(boundaries), 3)
        boundaries_to_check = [
            [2460619.0636363635, 2460619.084848485],
            [2460619.1666666665, 2460619.187878788],
            [2460619.2636363637, 2460619.2878787876]
        ]

        for each_boundary, boundary_to_check in zip(boundaries, boundaries_to_check):
            self.assertAlmostEquals(each_boundary.lower.jd, boundary_to_check[0])
            self.assertAlmostEquals(each_boundary.upper.jd, boundary_to_check[1])

    def test_boundaries_period(self):
        xlc = XLightCurve(time=self.time.tolist(), flux=self.flux.tolist())
        minimas = xlc.minima_mean()
        period = np.diff(minimas.time.jd).mean()
        t0 = minimas[0]
        boundaries = xlc.boundaries_period(t0.time, period)
        self.assertEqual(len(boundaries), 4)
        boundaries_to_check = [
            [2460619.0666340576, 2460619.086482542],
            [2460619.1658764817, 2460619.185724966],
            [2460619.265118906, 2460619.2849673904]
        ]
        for each_boundary, boundary_to_check in zip(boundaries, boundaries_to_check):
            self.assertAlmostEquals(each_boundary.lower.jd, boundary_to_check[0])
            self.assertAlmostEquals(each_boundary.upper.jd, boundary_to_check[1])

    def test_chop(self):
        xlc = XLightCurve(time=self.time.tolist(), flux=self.flux.tolist())
        chopped_xlc = xlc.chop()
        self.assertEqual(len(list(chopped_xlc)), 3)

    def test_smooth_b_spline(self):
        xlc = XLightCurve(time=self.time.tolist(), flux=self.flux.tolist())
        smooth_xlc = xlc.smooth_b_spline()
        self.assertNotEqual((smooth_xlc.flux - self.flux).value.mean(), 0)

    def test_smooth_savgol_filter(self):
        xlc = XLightCurve(time=self.time.tolist(), flux=self.flux.tolist())
        smooth_xlc = xlc.smooth_b_spline()
        self.assertNotEqual((smooth_xlc.flux - self.flux).value.mean(), 0)

    def test_smooth_butterworth_filter(self):
        xlc = XLightCurve(time=self.time.tolist(), flux=self.flux.tolist())
        smooth_xlc = xlc.smooth_butterworth_filter()
        self.assertNotEqual((smooth_xlc.flux - self.flux).value.mean(), 0)



from .lc import XLightCurve
from .sky import Sky
from .portal import Portal
from .utils import Boundaries
from .utils import Minima
from .catalgues import ExoplanetEU
from .catalgues import NASAExoplanetArchive
from .catalgues import Period

__all__ = [
    "XLightCurve", "Sky", "Portal", "Boundaries",
    "Minima", "ExoplanetEU", "NASAExoplanetArchive", "Period"
]

__version__ = "0.0.1 Beta"
__author__ = "Mohammad Niaei"
__license__ = "GNU/GPL V3"



from abc import ABC, abstractmethod
from logging import Logger
from typing import Self, Iterator, Optional, Callable, Any

from astropy.time import Time
from lightkurve import LightCurve

from .utils import Boundaries, Minima


class ModelXLightCurve(ABC):
    @classmethod
    @abstractmethod
    def from_lightcurve(cls, lightcurve: LightCurve, logger: Optional[Logger] = None):
        """Create an XLightCurve object from a LightCurve object"""

    @abstractmethod
    def boundaries_extrema(self, sigma_multiplier: float = 1.0) -> Boundaries:
        """Find boundaries of each minimum using curve fit"""

    @abstractmethod
    def boundaries_period(self, t0: Time, period: float, gap: float = 0.1) -> Boundaries:
        """Find boundaries of each minimum using period"""

    @abstractmethod
    def chop(self, boundaries: Optional[Boundaries] = None) -> Iterator[Self]:
        """Chops the XLightCurve according to boundaries"""

    @abstractmethod
    def smooth_b_spline(self, s: float = 1.0) -> Self:
        """Smooths the lightcurve using splrep"""

    @abstractmethod
    def smooth_savgol_filter(self, window: int = 21, order: int = 2) -> Self:
        """Smooths the lightcurve using savgol filter"""

    @abstractmethod
    def smooth_butterworth_filter(self, cutoff_freq: float = 0.5, sampling_rate: float = 10.0,
                                  order: int = 4) -> Self:
        """Smooths the lightcurve using butterworth filter"""

    @abstractmethod
    def minima_mean(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """"Returns minima times using mean value of flux in each minima"""

    @abstractmethod
    def minima_median(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns minima times using median value of flux in each minima"""

    @abstractmethod
    def minima_local(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns minima times using local minima value of flux in each minima"""

    @abstractmethod
    def minima_fit(self, boundaries: Optional[Boundaries] = None, deg: int = 2) -> Minima:
        """Returns minima times by fitting a polynomial and finding the root of it for each minima"""

    @abstractmethod
    def minima_periodogram(self) -> Minima:
        """Returns minima times using periodogram of each minima"""

    @abstractmethod
    def minima_kwee_van_woerden(self, number_of_folds: int = 5, time_off: bool = True, init_min_flux: bool = False,
                                boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns minima times using Kwee Van Woerden of each minima"""

    @abstractmethod
    def minima_bisector(self, middle_selector: Optional[Callable[[Any], float]] = None, number_of_chords: int = 5,
                        sigma_multiplier: float = 0.1, boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns minima times using bisector of each minima"""

    @abstractmethod
    def mid_egress_fielder(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns mid-egress time of each minima using Fielder method"""

    @abstractmethod
    def mid_ingress_fielder(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns mid-ingress time of each minima using Fielder method"""

    @abstractmethod
    def minima_fielder(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns minima times using Fielder method"""

    @abstractmethod
    def mid_egress_wood(self, boundaries: Optional[Boundaries] = None, median_width: int = 5,
                        avg_width: int = 5, egress_width: float = 10.0) -> Minima:
        """Returns mid-egress time of each minima using Wood method"""

    @abstractmethod
    def minima_thoroughgood(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """Returns minima times using Thoroughgood method"""



from unittest import TestCase

from octans import Sky, Period
from octans.errors import NoExoplanetEUObjectError, NoNASAExoplanetArchiveObjectError


class TestPeriod(TestCase):
    def test_create(self):
        wasp_3_b = Sky("wasp-3 b")
        _ = Period(wasp_3_b)

    def test_period_all(self):
        wasp_3_b = Sky("wasp-3 b")
        exoplanet = Period(wasp_3_b)
        for i in ["source", "P", "PEmin", "PEmax"]:
            self.assertIn(i, exoplanet.all().columns)

    def test_period_all_does_not_exist(self):
        xy_leo = Sky("xy leo")
        exoplanet = Period(xy_leo)
        with self.assertRaises((NoExoplanetEUObjectError, NoNASAExoplanetArchiveObjectError)):
            _ = exoplanet.all()

    def test_period_exoplanet_eu(self):
        wasp_3_b = Sky("wasp-3 b")
        exoplanet = Period(wasp_3_b)
        for i in ["P", "PEmin", "PEmax"]:
            self.assertIn(i, exoplanet.exoplanet_eu().columns)

    def test_period_exoplanet_eu_does_not_exist(self):
        xy_leo = Sky("xy leo")
        exoplanet = Period(xy_leo)
        with self.assertRaises(NoExoplanetEUObjectError):
            _ = exoplanet.exoplanet_eu()

    def test_period_nasa_exoplanet_archive(self):
        wasp_3_b = Sky("wasp-3 b")
        exoplanet = Period(wasp_3_b)
        for i in ["P", "PEmin", "PEmax"]:
            self.assertIn(i, exoplanet.nasa_exoplanet_archive().columns)

    def test_period_nasa_exoplanet_archive_does_not_exist(self):
        xy_leo = Sky("xy leo")
        exoplanet = Period(xy_leo)
        with self.assertRaises(NoNASAExoplanetArchiveObjectError):
            _ = exoplanet.nasa_exoplanet_archive()



from logging import getLogger, Logger

import pandas as pd
import requests
from typing import Optional

from astropy import units
from astropy.coordinates import SkyCoord
from tqdm import tqdm

from . import Sky
from .model_catalgues import ExoplanetEUModel, NASAExoplanetArchiveModel, PeriodModel
from .errors import NoExoplanetEUObjectError, NoNASAExoplanetArchiveObjectError
from .utils import database_dir, NAngleType, degree_checker

from astroquery.ipac.nexsci.nasa_exoplanet_archive import NasaExoplanetArchive


class ExoplanetEU(ExoplanetEUModel):
    def __init__(self, sky: Sky, max_dist: NAngleType = 20 * units.arcsec, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.database_file = database_dir() / "exoplanet_eu.csv"
        self.url = "https://exoplanet.eu/catalog/csv/"

        self.sky = sky
        self.max_dist = max_dist
        self.the_max_dist = degree_checker(self.max_dist)
        self.database = self.load()

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.sky})"

    def __repr__(self) -> str:
        return self.__str__()

    def exist(self) -> bool:
        return bool(self.database_file.exists())

    def download(self, force=False) -> None:
        if self.exist() and not force:
            self.logger.warning("Databse already exist. If you want to force download it use `force=True`")
            return

        with (open(self.database_file, "w", encoding="utf-8") as database,
              requests.get(self.url, stream=True) as response):
            response.raise_for_status()
            for chunk in tqdm(response.iter_content(chunk_size=1024)):
                if chunk:
                    database.write(chunk.decode("utf-8"))

    def load(self) -> pd.DataFrame:
        if not self.exist():
            self.download()

        try:
            return pd.read_csv(self.database_file)
        except Exception as e:
            self.logger.warning(e)

        self.download(force=True)
        return pd.read_csv(self.database_file)

    def resolve(self) -> pd.DataFrame:
        database = self.database[self.database.ra.notna() | self.database.dec.notna()]
        skycoord = SkyCoord(ra=database["ra"], dec=database["dec"], unit="deg")

        closest_ids, closest_dists, _ = skycoord.match_to_catalog_sky(SkyCoord([self.sky.skycoord]))
        mask = closest_dists < self.the_max_dist

        if mask.sum() == 0:
            raise NoExoplanetEUObjectError("There's no such ExoplanetEU object")

        data = database[mask]
        return data

    def period(self) -> pd.DataFrame:
        data = self.resolve()
        data = data[["name", "orbital_period", "orbital_period_error_min", "orbital_period_error_max"]]
        data.rename(columns={
            'orbital_period': 'P',
            'orbital_period_error_min': 'PEmin',
            'orbital_period_error_max': 'PEmax'
        }, inplace=True)
        data = data.reset_index()
        return data


class NASAExoplanetArchive(NASAExoplanetArchiveModel):
    def __init__(self, sky: Sky, max_dist: NAngleType = 20 * units.arcsec, logger: Optional[Logger] = None) -> None:
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.sky = sky
        self.max_dist = max_dist
        self.the_max_dist = degree_checker(self.max_dist)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.sky})"

    def __repr__(self) -> str:
        return self.__str__()

    def resolve(self) -> pd.DataFrame:
        table = NasaExoplanetArchive.query_region(table="pscomppars", coordinates=self.sky.skycoord,
                                                  radius=self.max_dist)
        closest_ids, closest_dists, _ = table["sky_coord"].match_to_catalog_sky(SkyCoord([self.sky.skycoord]))
        mask = closest_dists < self.the_max_dist

        if mask.sum() == 0:
            raise NoNASAExoplanetArchiveObjectError("There's no such ExoplanetEU object")

        data = table.to_pandas()[mask]
        return data

    def period(self) -> pd.DataFrame:
        data = self.resolve()
        data = data[["pl_name", "pl_orbper", "pl_orbpererr1", "pl_orbpererr2"]]
        data.rename(
            columns={
                'pl_name': 'name',
                'pl_orbper': 'P',
                'pl_orbpererr1': 'PEmin',
                'pl_orbpererr2': 'PEmax'
            },
            inplace=True)
        data = data.reset_index()
        return data


class Period(PeriodModel):
    def __init__(self, sky: Sky, max_dist: NAngleType = 20 * units.arcsec, logger: Optional[Logger] = None):
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        self.sky = sky
        self.max_dist = max_dist
        self.the_max_dist = degree_checker(self.max_dist)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.sky})"

    def __repr__(self) -> str:
        return self.__str__()

    def nasa_exoplanet_archive(self) -> pd.DataFrame:
        nasa = NASAExoplanetArchive(self.sky, max_dist=self.max_dist, logger=self.logger)
        return nasa.period()

    def exoplanet_eu(self) -> pd.DataFrame:
        eu = ExoplanetEU(self.sky, max_dist=self.max_dist, logger=self.logger)
        return eu.period()

    def all(self) -> pd.DataFrame:
        nasa = NASAExoplanetArchive(self.sky, max_dist=self.max_dist, logger=self.logger)
        nasa_data = nasa.period()
        nasa_data["source"] = "NEA"
        eu = ExoplanetEU(self.sky, max_dist=self.max_dist, logger=self.logger)
        eu_data = eu.period()
        eu_data["source"] = "EEU"
        data = pd.concat([nasa_data, eu_data], ignore_index=True)
        del data["index"]
        return data[["source", "P", "PEmin", "PEmax"]]



class ObjectNotFoundError(Exception):
    """Raised when an Object not found"""


class NoExoplanetEUObjectError(ObjectNotFoundError):
    """Raised when an ExoplanetEU object not found"""


class NoNASAExoplanetArchiveObjectError(ObjectNotFoundError):
    """Raised when an ExoplanetEU object not found"""


class NoASASObjectError(ObjectNotFoundError):
    """Raised when an ASAS object not found"""


class NoVarAstroObjectError(ObjectNotFoundError):
    """Raised when an ASAS object not found"""


class NoDataFoundError(Exception):
    """Raised when no data was found"""


class PageNotFoundError(Exception):
    """Raised when page not found"""


class NoLightCurveError(NoDataFoundError):
    """Raised when a light curve not found"""



from astropy.coordinates import SkyCoord
from astropy.coordinates.name_resolve import NameResolveError

from octans import Sky, Portal
from unittest import TestCase

from octans.errors import NoLightCurveError, NoASASObjectError, NoVarAstroObjectError


class TestSky(TestCase):

    def test_create(self):
        sky = Sky("xy leo")
        portal = Portal(sky)
        sky_coord = SkyCoord.from_name("xy Leo")
        self.assertEqual(portal.sky.skycoord, sky_coord)

    def test_create_from_name(self):
        sky = Sky("xy leo")
        portal = Portal.from_name("xy leo")
        self.assertEqual(sky.skycoord, portal.sky.skycoord)

    def test_create_from_name_does_not_exist(self):
        with self.assertRaises(NameResolveError):
            _ = Portal.from_name("wa dra")

    def test_change_sky_coords(self):
        portal = Portal.from_name("xy leo")
        new_sky = Sky("aw dra")
        with self.assertRaises(AttributeError):
            portal.sky = new_sky

    def test_kepler(self):
        portal = Portal.from_name("kepler9")
        kepler = portal.kepler()
        self.assertEqual(len(list(kepler)), 50)

    def test_kepler_does_not_exist(self):
        portal = Portal.from_name("wasp 3")
        kepler = portal.kepler()
        with self.assertRaises(NoLightCurveError):
            _ = list(kepler)

    def test_tess(self):
        portal = Portal.from_name("kepler9")
        tess = portal.tess()
        self.assertEqual(len(list(tess)), 14)

    def test_tess_does_not_exist(self):
        portal = Portal.from_name("Alpha pegasi")
        tess = portal.tess()
        with self.assertRaises(NoLightCurveError):
            _ = list(tess)

    def test_asas(self):
        portal = Portal.from_name("xy leo")
        asas = portal.asas()
        self.assertEqual(len(list(asas)), 5)

    def test_asas_does_not_exist(self):
        portal = Portal.from_name("aw dra")
        with self.assertRaises(NoASASObjectError):
            _ = portal.asas()

    def test_var_astro(self):
        portal = Portal.from_name("xy leo")
        var_astro = portal.var_astro()
        self.assertEqual(len(var_astro), 317)

    def test_var_astro_does_not_exist(self):
        portal = Portal.from_name("aw dra")
        with self.assertRaises(NoLightCurveError):
            _ = portal.var_astro()

    def test_etd(self):
        portal = Portal.from_name("wasp 2")
        etd = portal.etd()
        self.assertEqual(len(etd), 199)

    def test_etd_does_not_exist(self):
        portal = Portal.from_name("xy leo")
        with self.assertRaises(NoVarAstroObjectError):
            _ = portal.etd()



from logging import Logger, getLogger
from typing import Iterator, Optional, Callable, Any

import numpy as np
from numpy import polyfit
from scipy.interpolate import splrep, splev, UnivariateSpline
from scipy.signal import argrelextrema, savgol_filter, butter, filtfilt, medfilt
from typing_extensions import Self

from lightkurve import LightCurve
from astropy.time import Time, TimeDelta

from .kvw import kvw
from .lc_model import ModelXLightCurve
from .utils import Boundaries, Minima, interpolate, normalize, neighbor, TimeType


class XLightCurve(ModelXLightCurve, LightCurve):
    def __init__(self, *args, logger: Optional[Logger] = None, **kwargs):
        if logger is None:
            self.logger = getLogger(__name__)
        else:
            self.logger = logger

        super().__init__(*args, **kwargs)

    @classmethod
    def from_lightcurve(cls, lightcurve: LightCurve, logger: Optional[Logger] = None):
        return cls(
            time=lightcurve.time, flux=lightcurve.flux, flux_err=lightcurve.flux_err, logger=logger
        )

    def boundaries_extrema(self, sigma_multiplier: float = 1.0) -> Boundaries:
        flux_std = self.flux.std()
        flux_mean = self.flux.mean()
        times = self.time[self.flux < flux_mean - sigma_multiplier * flux_std]
        right_diffs = np.diff(times.jd)
        right_indices = np.concatenate([
            np.where(right_diffs > right_diffs.mean())[0],
            np.array([len(times) - 1])
        ])

        left_diffs = abs(np.diff(times.jd[::-1]))
        left_indices = len(times) - 1 - np.concatenate([
            np.array([len(times) - 1]),
            np.where(left_diffs > left_diffs.mean())[0]
        ])

        right_indices.sort()
        right_time = Time(
            [
                times[ri].jd
                for ri in right_indices
            ],
            format='jd'
        )

        left_indices.sort()
        left_time = Time(
            [
                times[li].jd
                for li in left_indices
            ],
            format='jd'
        )

        return Boundaries(
            left_time, right_time,
        )

    def boundaries_period(self, t0: Time, period: float, gap: float = 0.1) -> Boundaries:
        if gap >= 0.5:
            raise ValueError("Gap must be greater than 0.5")

        time_min = self.time.min()
        time_max = self.time.max()
        diff = (t0 - time_min).jd % period
        minimas = np.arange(time_min.jd + diff, time_max.jd + diff, period)

        left_time = Time(
            [
                minima - gap * period
                for minima in minimas
            ],
            format='jd'
        )

        right_time = Time(
            [
                minima + gap * period
                for minima in minimas
            ],
            format='jd'
        )

        return Boundaries(left_time, right_time)

    def chop(self, boundaries: Optional[Boundaries] = None) -> Iterator[Self]:

        if boundaries is None:
            boundaries_to_use = self.boundaries_extrema()
        else:
            boundaries_to_use = boundaries

        for boundary in boundaries_to_use:
            yield self[np.logical_and(self.time.jd > boundary.lower.jd, self.time.jd < boundary.upper.jd)]

    def smooth_b_spline(self, s: float = 1.0) -> Self:
        fluxes = np.array(self.flux.value)
        flux_mask = np.isnan(fluxes)

        bspl = splrep(self.time.jd[~flux_mask], np.array(self.flux.value)[~flux_mask], s=s)

        xlc = self.__class__(
            time=self.time[~flux_mask],
            flux=splev(self.time.jd[~flux_mask], bspl) * self.flux.unit,
            flux_err=self.flux_err[~flux_mask]
        )
        return xlc

    def smooth_savgol_filter(self, window: int = 21, order: int = 2) -> Self:
        fluxes = np.array(self.flux.value)
        flux_mask = np.isnan(fluxes)

        xlc = self.__class__(
            time=self.time[~flux_mask],
            flux=savgol_filter(np.array(self.flux.value)[~flux_mask], window, order) * self.flux.unit,
            flux_err=self.flux_err[~flux_mask]
        )

        return xlc

    # https://www.delftstack.com/howto/python/smooth-data-in-python/
    def smooth_butterworth_filter(self, cutoff_freq: float = 0.5, sampling_rate: float = 10.0,
                                  order: int = 4) -> Self:
        fluxes = np.array(self.flux.value)
        flux_mask = np.isnan(fluxes)

        nyquist = 0.5 * sampling_rate
        normal_cutoff = cutoff_freq / nyquist
        ba = butter(order, normal_cutoff, btype="low", analog=False)

        xlc = self.__class__(
            time=self.time[~flux_mask],
            flux=filtfilt(*ba, np.array(self.flux.value)[~flux_mask]) * self.flux.unit,
            flux_err=self.flux_err[~flux_mask]
        )
        return xlc

    def minima_mean(self, boundaries: Optional[Boundaries] = None) -> Minima:
        times = []
        time_error = []
        for chopped in self.chop(boundaries):
            try:
                mean_time = np.mean(chopped.time.jd)
                std_time = np.std(chopped.time.jd)
                if not np.isnan(mean_time):
                    times.append(mean_time)
                    time_error.append(std_time)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            format='jd'
        )

        error = TimeDelta(
            time_error, format='jd'
        )
        return Minima(times, error)

    def minima_median(self, boundaries: Optional[Boundaries] = None) -> Minima:
        times = []
        time_error = []
        for chopped in self.chop(boundaries):
            try:
                median_time = np.median(chopped.time.jd)
                if not np.isnan(median_time):
                    times.append(median_time)
                    q1 = np.percentile(chopped.time.jd, 25)
                    q3 = np.percentile(chopped.time.jd, 75)
                    time_error.append(max(abs(q1 - median_time), abs(q3 - median_time)))
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def minima_local(self, boundaries: Optional[Boundaries] = None) -> Minima:
        times = []
        time_error = []
        for chopped in self.chop(boundaries):
            try:
                local_minimas = argrelextrema(chopped.flux.value, np.less)
                if len(local_minimas) == 0:
                    self.logger.warning("No local minima was found")
                    continue

                local_minima = local_minimas[0]

                if len(local_minima) == 0:
                    self.logger.warning("No local minima was found")
                    continue

                times.append(chopped.time[local_minimas].jd[0])
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def minima_fit(self, boundaries: Optional[Boundaries] = None, deg: int = 2) -> Minima:
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                if len(chopped) <= 2:
                    continue
                coefficients, (res,), *_ = polyfit(chopped.time.jd, chopped.flux.value, deg=deg, full=True)
                derivative_coefficients = np.polyder(coefficients)
                roots = np.roots(derivative_coefficients)

                times.append(roots[0])
                time_error.append(res)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def minima_periodogram(self) -> Minima:

        pg = self.normalize(unit='ppm').to_periodogram()
        folded_xlc = self.fold(pg.period_at_max_power)

        first_minimum = folded_xlc.time[folded_xlc.flux == folded_xlc.flux.min()][0].jd + self.time.min().jd
        times = []
        time_error = []
        while first_minimum < self.time.max().jd:
            try:
                first_minimum += pg.period_at_max_power.value
                times.append(first_minimum)
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    # https://github.com/hdeeg/KvW
    def minima_kwee_van_woerden(self, number_of_folds: int = 5, time_off: bool = True, init_min_flux: bool = False,
                                boundaries: Optional[Boundaries] = None) -> Minima:
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                minima, minima_err, kvw_err, flag = kvw(
                    chopped.time.jd, chopped.flux.value,
                    nfold=number_of_folds, notimeoff=not time_off,
                    init_minflux=init_min_flux, rms=self.flux_err.value.mean()
                )
                times.append(minima)
                time_error.append(kvw_err)

            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def minima_bisector(self, middle_selector: Optional[Callable[[Any], float]] = None, number_of_chords: int = 5,
                        sigma_multiplier: float = 0.1, boundaries: Optional[Boundaries] = None) -> Minima:
        """Find minima times of each boundary using the bisector method"""
        if middle_selector is None:
            middle_selector = np.mean

        if number_of_chords < 3:
            raise ValueError("number of chord must be greater than 3")

        times = []
        time_error = []
        for chopped in self.chop(boundaries):
            try:
                std_time = np.std(chopped.time.jd)

                arg_middle = neighbor(chopped.time.jd, middle_selector(chopped.time.jd))

                flux_argmin = np.argmin(chopped.flux)
                flux_min = chopped.flux[flux_argmin]
                flux_max = min(chopped[:arg_middle].flux.max(), chopped[arg_middle:].flux.max())
                flux_std = chopped.flux.std()
                flux_lines = np.linspace(flux_min + flux_std * sigma_multiplier,
                                         flux_max - flux_std * sigma_multiplier, number_of_chords + 2)[1:-1]
                points = []
                for the_flux in flux_lines:
                    left_signs = np.sign(chopped[:arg_middle].flux - the_flux)
                    left_sign_change_index = ((np.roll(left_signs, 1) - left_signs) != 0).astype(int)[1:].argmax()

                    right_signs = np.sign(chopped[arg_middle:].flux - the_flux)
                    right_sign_change_index = ((np.roll(right_signs, 1) - right_signs) != 0).astype(int)[1:].argmax()

                    left_point_1_flux = chopped[:arg_middle].flux[left_sign_change_index]
                    left_point_2_flux = chopped[:arg_middle].flux[left_sign_change_index + 1]
                    right_point_1_flux = chopped[arg_middle:].flux[right_sign_change_index]
                    right_point_2_flux = chopped[arg_middle:].flux[right_sign_change_index + 1]
                    left_flux_propagation = normalize(the_flux, left_point_1_flux, left_point_2_flux)
                    right_flux_propagation = normalize(the_flux, right_point_1_flux, right_point_2_flux)

                    left_point_1_time = chopped[:arg_middle].time[left_sign_change_index].jd
                    left_point_2_time = chopped[:arg_middle].time[left_sign_change_index + 1].jd
                    right_point_1_time = chopped[arg_middle:].time[right_sign_change_index].jd
                    right_point_2_time = chopped[arg_middle:].time[right_sign_change_index + 1].jd

                    left_time = interpolate(left_flux_propagation, left_point_1_time, left_point_2_time)
                    right_time = interpolate(right_flux_propagation, right_point_1_time, right_point_2_time)

                    points.append((left_time + right_time) / 2)

                times.append(np.mean(points))
                time_error.append(std_time)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def mid_egress_fielder(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """
        First method of estimating egress of the WD is the spline fitting derivative method [e.g.,
            Fiedler et al., 1997]. This method is depicted in Figure 2 and individual steps are:
            1. Points of light curve are smoothed by an approximative cubic spline function.
            2. Derivation of the approximative cubic spline function.
            3. The time of the mid-egress is time of maximum of the derivative.
        """
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                spline = UnivariateSpline(chopped.time.jd, chopped.flux.value, s=0, k=3)
                derivative = spline.derivative()
                derivative_times = np.linspace(chopped.time.jd.min(), chopped.time.jd.max(), 1000)
                derivative_values = derivative(derivative_times)
                egress_time = derivative_times[np.argmax(derivative_values)]

                times.append(egress_time)
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def mid_ingress_fielder(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """
        First method of estimating egress of the WD is the spline fitting derivative method [e.g.,
            Fiedler et al., 1997]. This method is depicted in Figure 2 and individual steps are:
            1. Points of light curve are smoothed by an approximative cubic spline function.
            2. Derivation of the approximative cubic spline function.
            3. The time of the mid-egress is time of minimum of the derivative.
        """
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                spline = UnivariateSpline(chopped.time.jd, chopped.flux.value, s=0, k=3)
                derivative = spline.derivative()
                derivative_times = np.linspace(chopped.time.jd.min(), chopped.time.jd.max(), 1000)
                derivative_values = derivative(derivative_times)
                ingress_time = derivative_times[np.argmin(derivative_values)]

                times.append(ingress_time)
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def minima_fielder(self, boundaries: Optional[Boundaries] = None) -> Minima:
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                spline = UnivariateSpline(chopped.time.jd, chopped.flux.value, s=0, k=3)
                derivative = spline.derivative()
                derivative_times = np.linspace(chopped.time.jd.min(), chopped.time.jd.max(), 1000)
                derivative_values = derivative(derivative_times)
                egress_time = derivative_times[np.argmin(derivative_values)]
                ingress_time = derivative_times[np.argmax(derivative_values)]

                times.append((egress_time + ingress_time) / 2)
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def mid_egress_wood(self, boundaries: Optional[Boundaries] = None, median_width: int = 5,
                        avg_width: int = 5, egress_width: float = 10.0) -> Minima:
        """
        Second method was proposed by Wood et al. [1985]. Method uses median and average filter
            for smoothing of the points of the light curve. Median filter of a given width changes each point
            of the light curve to a median of a surrounding points of the given width and average filter
            changes each point on a average value of the points in a surrounding of the given width. This
            method is depicted in Figure 3 and individual steps are:
            1. Points of light curve are smoothed by median filter. This step can be omitted, if the
            measurements have small uncertainties.
            2. Derivative of smoothed light curve.
            3. Using average filter of width equal to the expected duration of egress.
            4. A spline function is fitted to points that are not lying within expected egress.
            5. Times of the beginning and the end of the egress of the WD are the same as times where
            the derivative differs greatly from the spline function.
            6. The time of the mid-egress of the WD is average value of these two values.
        """
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                smoothed_curve = medfilt(chopped.flux.value, kernel_size=median_width)
                if len(smoothed_curve) < 3:
                    self.logger.warning("Could not find enough points for egress estimation")
                    continue

                derivative = np.gradient(smoothed_curve, chopped.time.jd)
                smoothed_derivative = savgol_filter(derivative, window_length=avg_width, polyorder=2)
                egress_indices = np.where(np.abs(smoothed_derivative) < egress_width)[0]
                spline_fit = UnivariateSpline(
                    chopped.time.jd[egress_indices], smoothed_derivative[egress_indices], s=0
                )
                diff = np.abs(smoothed_derivative - spline_fit(chopped.time.jd))
                significant_diff_indices = np.where(diff > np.percentile(diff, 95))[0]

                if len(significant_diff_indices) < 2:
                    self.logger.warning("Could not find enough significant points for egress estimation")
                    continue

                start_egress_time = chopped.time.jd[significant_diff_indices[0]]
                end_egress_time = chopped.time.jd[significant_diff_indices[-1]]

                mid_egress_time = (start_egress_time + end_egress_time) / 2

                times.append(mid_egress_time)
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def mid_ingress_wood(self, boundaries: Optional[Boundaries] = None, median_width: int = 5,
                         avg_width: int = 5, egress_width: float = 10.0) -> Minima:
        """
        Second method was proposed by Wood et al. [1985]. Method uses median and average filter
            for smoothing of the points of the light curve. Median filter of a given width changes each point
            of the light curve to a median of a surrounding points of the given width and average filter
            changes each point on a average value of the points in a surrounding of the given width. This
            method is depicted in Figure 3 and individual steps are:
            1. Points of light curve are smoothed by median filter. This step can be omitted, if the
            measurements have small uncertainties.
            2. Derivative of smoothed light curve.
            3. Using average filter of width equal to the expected duration of egress.
            4. A spline function is fitted to points that are not lying within expected egress.
            5. Times of the beginning and the end of the egress of the WD are the same as times where
            the derivative differs greatly from the spline function.
            6. The time of the mid-egress of the WD is average value of these two values.
        """
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                smoothed_curve = medfilt(chopped.flux.value, kernel_size=median_width)
                if len(smoothed_curve) < 3:
                    self.logger.warning("Could not find enough points for egress estimation")
                    continue

                derivative = np.gradient(smoothed_curve, chopped.time.jd)
                smoothed_derivative = savgol_filter(derivative, window_length=avg_width, polyorder=2)
                ingress_indices = np.where(np.abs(smoothed_derivative) > egress_width)[0]
                spline_fit = UnivariateSpline(
                    chopped.time.jd[ingress_indices], smoothed_derivative[ingress_indices], s=0
                )
                diff = np.abs(smoothed_derivative - spline_fit(chopped.time.jd))
                significant_diff_indices = np.where(diff > np.percentile(diff, 95))[0]

                if len(significant_diff_indices) < 2:
                    self.logger.warning("Could not find enough significant points for ingress estimation")
                    continue

                start_ingress_time = chopped.time.jd[significant_diff_indices[0]]
                end_ingress_time = chopped.time.jd[significant_diff_indices[-1]]

                mid_ingress_time = (start_ingress_time + end_ingress_time) / 2

                times.append(mid_ingress_time)
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def minima_thoroughgood(self, boundaries: Optional[Boundaries] = None) -> Minima:
        """
        The accretion disc outshines the WD and the bright spot. The egress of the WD can be
            hardly identified and therefore derivative methods are not usable. Precise time of the mid-
            eclipse is hard to estimate because of asymmetrical shape of the eclipse. Many authors use a
            parabola fitting method [e.g. Thoroughgood et al., 2004] to the eclipse light curve or to the
            lower half of the eclipse curve. It is important to use same method for all estimations of the
            mid-eclipse times of each nova or nova-like variable to ensure smaller systematic errors. The
            parabola fitting method is depicted on the left part of Figure 4 and individual steps are:
            1. Fitting parabola to the eclipse part of the light curve (or to the lower half of the eclipse
            part of the light curve).
            2. The time of the mid-eclipse is minimum of the parabola.
        """
        times = []
        time_error = []

        for chopped in self.chop(boundaries):
            try:
                coefficients = np.polyfit(chopped.time.jd, chopped.flux.value, 2)
                a, b, _ = coefficients
                x_min = -b / (2 * a)
                times.append(x_min)
                time_error.append(np.nan)
            except Exception as e:
                self.logger.warning(e)

        times = Time(
            times,
            scale=self.time.scale, format=self.time.format
        )

        error = TimeDelta(
            time_error, format='jd'
        )

        return Minima(times, error)

    def period(self) -> TimeType:
        pg = self.to_periodogram()
        return pg.period_at_max_power
